@@ -308,6 +308,7 @@ public:
 		_refnames(refnames),
 		_numWrappers(0),
 		_locks(),
+		ts_wrap(NULL),
 		INIT_HIT_DUMPS,
 		onePairFile_(onePairFile),
 		sampleMax_(sampleMax),
@@ -321,9 +322,8 @@ public:
 		ssmode_(ios_base::out)
 	{
 		_outs.push_back(out);
-                vector<MUTEX_T*>::iterator it;
-                
-                _locks.push_back(new MUTEX_T);
+		vector<MUTEX_T*>::iterator it;
+		_locks.push_back(new MUTEX_T);
 		initDumps();
 	}
 
@@ -381,9 +381,8 @@ public:
 	 * lock or any of the per-stream locks will be contended.
 	 */
 	void addWrapper() {
-            numWrapper_mutex_m.lock();
-            _numWrappers++;
-            numWrapper_mutex_m.unlock();
+		ThreadSafe ts(&numWrapper_mutex_m);
+		_numWrappers++;
 	}
 
 	/**
@@ -419,25 +418,27 @@ public:
 		if(_outs.size() > 1 && end-start > 2) {
 			sort(hs.begin() + start, hs.begin() + end);
 		}
-		char buf[4096];
-		for(size_t i = start; i < end; i++) {
-			const Hit& h = hs[i];
-			assert(h.repOk());
-			bool diff = false;
-			if(i > start) {
-				diff = (refIdxToStreamIdx(h.h.first) != refIdxToStreamIdx(hs[i-1].h.first));
-				if(diff) unlock(hs[i-1].h.first);
+		string buf(4096, (char) 0);
+		ostringstream ss(ssmode_);
+		size_t i = start;
+		while(i < end) {
+			size_t strIdx = refIdxToStreamIdx(hs[i].h.first);
+			{
+				do {
+					assert(hs[i].repOk());
+					append(ss, hs[i]);
+					{
+						ThreadSafe _ts(_locks[strIdx]);
+						out(hs[i].h.first).writeChars(ss.str().c_str(), ss.tellp());
+					}
+					ss.seekp(0);
+					ss.clear();
+					i++;
+				} while(refIdxToStreamIdx(hs[i].h.first) == strIdx && i < end);
 			}
-			ostringstream ss(ssmode_);
-			ss.rdbuf()->pubsetbuf(buf, 4096);
-			append(ss, h);
-			if(i == start || diff) {
-				lock(h.h.first);
-			}
-			out(h.h.first).writeChars(buf, ss.tellp());
 		}
-		unlock(hs[end-1].h.first);
-		GUARD_LOCK(main_mutex_m);
+		
+        ThreadSafe _ts(&main_mutex_m);
 		commitHits(hs);
 		first_ = false;
 		numAligned_++;
@@ -550,22 +551,6 @@ public:
 		return *(_outs[strIdx]);
 	}
 
-	/**
-	 * Lock the monolithic lock for this HitSink.  This is useful when,
-	 * for example, outputting a read to an unaligned-read file.
-	 */
-	void mainlock() {
-                main_mutex_m.lock();
-	}
-
-	/**
-	 * Unlock the monolithic lock for this HitSink.  This is useful
-	 * when, for example, outputting a read to an unaligned-read file.
-	 */
-	void mainunlock() {
-                main_mutex_m.unlock();
-	}
-
 	/**
 	 * Return true iff this HitSink dumps aligned reads to an output
 	 * stream (i.e., iff --alfa or --alfq are specified).
@@ -606,10 +591,11 @@ public:
 	 */
 	void dumpAlign(PatternSourcePerThread& p) {
 		if(!dumpAlignFlag_) return;
-		if(!p.paired() || onePairFile_) {
+		const bool paired = p.bufa().mate > 0;
+		if(!paired || onePairFile_) {
 			// Dump unpaired read to an aligned-read file of the same format
 			if(!dumpAlBase_.empty()) {
-				GUARD_LOCK(dumpAlignLock_);
+				ThreadSafe _ts(&dumpAlignLock_);
 				if(dumpAl_ == NULL) {
 					assert(dumpAlQv_ == NULL);
 					dumpAl_ = openOf(dumpAlBase_, 0, "");
@@ -628,7 +614,7 @@ public:
 			// Dump paired-end read to an aligned-read file (or pair of
 			// files) of the same format
 			if(!dumpAlBase_.empty()) {
-				GUARD_LOCK(dumpAlignLockPE_);
+				ThreadSafe _ts(&dumpAlignLockPE_);
 				if(dumpAl_1_ == NULL) {
 					assert(dumpAlQv_1_ == NULL);
 					assert(dumpAlQv_2_ == NULL);
@@ -660,10 +646,11 @@ public:
 	 */
 	void dumpUnal(PatternSourcePerThread& p) {
 		if(!dumpUnalignFlag_) return;
-		if(!p.paired() || onePairFile_) {
+		const bool paired = p.bufa().mate > 0;
+		if(!paired || onePairFile_) {
 			// Dump unpaired read to an unaligned-read file of the same format
 			if(!dumpUnalBase_.empty()) {
-				GUARD_LOCK(dumpUnalLock_);
+				ThreadSafe _ts(&dumpUnalLock_);
 				if(dumpUnal_ == NULL) {
 					assert(dumpUnalQv_ == NULL);
 					dumpUnal_ = openOf(dumpUnalBase_, 0, "");
@@ -682,7 +669,7 @@ public:
 			// Dump paired-end read to an unaligned-read file (or pair
 			// of files) of the same format
 			if(!dumpUnalBase_.empty()) {
-				GUARD_LOCK(dumpUnalLockPE_);
+				ThreadSafe _ts(&dumpUnalLockPE_);
 				if(dumpUnal_1_ == NULL) {
 					assert(dumpUnal_1_ == NULL);
 					assert(dumpUnal_2_ == NULL);
@@ -715,10 +702,11 @@ public:
 			if(dumpUnalignFlag_) dumpUnal(p);
 			return;
 		}
-		if(!p.paired() || onePairFile_) {
+		const bool paired = p.bufa().mate > 0;
+		if(!paired || onePairFile_) {
 			// Dump unpaired read to an maxed-out-read file of the same format
 			if(!dumpMaxBase_.empty()) {
-				GUARD_LOCK(dumpMaxLock_);
+				ThreadSafe _ts(&dumpMaxLock_);
 				if(dumpMax_ == NULL) {
 					dumpMax_ = openOf(dumpMaxBase_, 0, "");
 					assert(dumpMax_ != NULL);
@@ -735,7 +723,7 @@ public:
 			// Dump paired-end read to a maxed-out-read file (or pair
 			// of files) of the same format
 			if(!dumpMaxBase_.empty()) {
-				GUARD_LOCK(dumpMaxLockPE_);
+				ThreadSafe _ts(&dumpMaxLockPE_);
 				if(dumpMax_1_ == NULL) {
 					assert(dumpMaxQv_1_ == NULL);
 					assert(dumpMaxQv_2_ == NULL);
@@ -763,7 +751,7 @@ public:
 	 * want to print a placeholder when output is chained.
 	 */
 	virtual void reportMaxed(vector<Hit>& hs, PatternSourcePerThread& p) {
-		GUARD_LOCK(main_mutex_m);
+		ThreadSafe _ts(&main_mutex_m);
 		numMaxed_++;
 	}
 
@@ -772,7 +760,7 @@ public:
 	 * want to print a placeholder when output is chained.
 	 */
 	virtual void reportUnaligned(PatternSourcePerThread& p) {
-		GUARD_LOCK(main_mutex_m);
+		ThreadSafe _ts(&main_mutex_m);
 		numUnaligned_++;
 	}
 
@@ -781,7 +769,7 @@ protected:
 	/// Implementation of hit-report
 	virtual void reportHit(const Hit& h) {
 		assert(h.repOk());
-		GUARD_LOCK(main_mutex_m);
+		ThreadSafe _ts(&main_mutex_m);
 		commitHit(h);
 		first_ = false;
 		if(h.mate > 0) numReportedPaired_++;
@@ -801,28 +789,6 @@ protected:
 		}
 	}
 
-	/**
-	 * Lock the output buffer for the output stream for reference with
-	 * index 'refIdx'.  By default, hits for all references are
-	 * directed to the same output stream, but if --refout is
-	 * specified, each reference has its own reference stream.
-	 */
-	void lock(size_t refIdx) {
-		size_t strIdx = refIdxToStreamIdx(refIdx);
-		_locks[strIdx]->lock();
-	}
-
-	/**
-	 * Lock the output buffer for the output stream for reference with
-	 * index 'refIdx'.  By default, hits for all references are
-	 * directed to the same output stream, but if --refout is
-	 * specified, each reference has its own reference stream.
-	 */
-	void unlock(size_t refIdx) {
-		size_t strIdx = refIdxToStreamIdx(refIdx);
-		_locks[strIdx]->unlock();
-	}
-
 	vector<OutFileBuf*> _outs;        /// the alignment output stream(s)
 	bool                _deleteOuts;  /// Whether to delete elements of _outs upon exit
 	RecalTable         *recalTable_;  /// recalibration table
@@ -831,6 +797,7 @@ protected:
 	vector<MUTEX_T*>     _locks;       /// pthreads mutexes for per-file critical sections
 	MUTEX_T             main_mutex_m;    /// pthreads mutexes for fields of this object
 	MUTEX_T             numWrapper_mutex_m;
+	ThreadSafe*         ts_wrap;      /// for mutual exclusion
 
 	// Output filenames for dumping
 	std::string dumpAlBase_;
@@ -1536,9 +1503,9 @@ protected:
 		HitSink::reportHit(h);
 		ostringstream ss;
 		append(ss, h);
-		lock(h.h.first);
+		
+		ThreadSafe _ts(_locks[refIdxToStreamIdx(h.h.first)]);
 		out(h.h.first).writeString(ss.str());
-		unlock(h.h.first);
 	}
 
 private:
@@ -1672,10 +1639,10 @@ protected:
 		if(count) HitSink::reportHit(h);
 		ostringstream ss;
 		append(ss, h);
+
 		// Make sure to grab lock before writing to output stream
-		lock(h.h.first);
+		ThreadSafe _ts(_locks[refIdxToStreamIdx(h.h.first)]);
 		out(h.h.first).writeString(ss.str());
-		unlock(h.h.first);
 	}
 
 private:
